use anyhow::Result;
use std::any::{Any, TypeId};

use crate::{world::ComponentChange, *};

enum CommandsTypes {
    InsertEntity(EntityObject),
    RemoveEntity(Entity),
    UpdateEntity(Entity, ComponentChange),

    Event(TypeId, Box<dyn Any + Send + Sync>),
}

#[derive(Default)]
pub struct Commands {
    commands: Vec<CommandsTypes>,
}

impl Commands {
    pub fn insert_entity(&mut self, entity_object: EntityObject) -> Entity {
        let entity = entity_object.as_entity();

        self.commands
            .push(CommandsTypes::InsertEntity(entity_object));

        entity
    }

    pub fn remove_entity(&mut self, entity: Entity) {
        self.commands.push(CommandsTypes::RemoveEntity(entity));
    }

    pub fn insert_component<T: EntityComponent>(&mut self, entity: Entity, component: T) {
        self.commands.push(CommandsTypes::UpdateEntity(
            entity,
            ComponentChange::Added(TypeId::of::<T>(), Box::new(component)),
        ));
    }

    pub fn remove_component<T: EntityComponent>(&mut self, entity: Entity) {
        self.commands.push(CommandsTypes::UpdateEntity(
            entity,
            ComponentChange::Removed(TypeId::of::<T>()),
        ));
    }

    pub fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.commands
            .push(CommandsTypes::Event(TypeId::of::<T>(), Box::new(payload)));
    }

    pub(crate) fn apply_deferred(self, world: &mut World) -> Result<()> {
        for command in self.commands {
            match command {
                CommandsTypes::InsertEntity(entity_object) => {
                    world.add_entity(entity_object)?;
                }
                CommandsTypes::RemoveEntity(entity) => world.remove_entity(entity),
                CommandsTypes::UpdateEntity(entity, component_change) => {
                    world.component_change(entity, component_change)
                }
                CommandsTypes::Event(type_id, any) => world.events.write_payload(type_id, any),
            }
        }

        Ok(())
    }
}
